/* This file contains the functions actually reqd. to carry out transactions.
 * The quoting is based on a risk kind of method.
 * You can quote to buy without having enough money and
 * quote for sale without having the stocks.
 * The validation is done at actual transaction time, triggering violations
 * in case of defaulters.
 * a system of loans will be soon put in.
 */

#include "stock.h"
#include "threads.h"

#define BUYER   1
#define SELLER  2

static qty_t buy_really (plid_t plid, cid_t cid, qty_t qty, pr_t pr,
			 int caller);

inline static void violation (plid_t plid);
inline static void set_last_trade (cid_t cid, pr_t pr);

inline static void set_last_trade (cid_t cid, pr_t pr) {
   /* Set the last_trade in the company history */
   company_t *c = &company[cid];
   
   c->last_trade = ++c->last_trade % max_history;
   c->trade_history[c->last_trade] = pr;
}

inline static void violation (plid_t plid) {
   /* This could be a cash or a qty violation..
    * punishment involves unable to trade for a few seconds
    */
   player_t *p = &player[plid];
   
#ifdef LOG
   fprintf (logfile, "Violation by player : %d\n", plid);
#endif
   
#ifdef DEBUG
   printf ("Violation by player : %d\n", plid);
#endif

   /*   pthread_kill (*p->sync_th, SIGALRM); */
   close (p->client);   
   /* not tested... this should trigger a kick_player () though */
   return;
}
   
/* buy_really () function doesn't take care of checking if reqd. qty is bought.
 * it just returns the qty actually bought. The way the transaction happens
 * is as follows :
 * 
 * 0) Check caller
 * 1) Check buy qty.
 * 2) Check cash.
 * 3) Check sell qty.
 * 4) Match buy and sell qtys.
 * 5) Change the last_traded price to show a new transaction.
 * 6) Give cash from buyer to seller.
 * 7) Give stocks from seller to buyer.
 * 8) Return the qty traded.
 * 
 * the caller is either set to BUYER or SELLER to decide the values of
 * variables such as plid and qty..
 * 
 * A -ve return indicates defaulting on part of the other chap, and hence
 * we need to signal the caller to come back again...
 */

static int buy_really (plid_t plid, cid_t cid, qty_t qty, pr_t pr, 
		       int caller) {
   
   stock_player_t *holding_buyer, *holding_seller;
   company_t *comp = &company[cid];
   cash_t bcash;    /* buyer's cash */
   cash_t cash_changing_hands;
   plid_t splid;
   qty_t sqty;  /* seller's qty */
        
   if (caller == BUYER) {
      splid = comp->sell[0]->plid;
      sqty = comp->sell[0]->qty;
   } else {
      sqty = qty;
      qty = comp->buy[0]->qty;
      splid = plid;
      plid = comp->buy[0]->plid;
   }
   
   /* Now we check if the seller has the qty he says he wants to sell..*/
   if ((holding_seller = holding_search (splid, cid)) == NULL) {
      sell_pr_remove (cid, splid, 0);
      violation (splid);
      return -1;
   }
      
   /* Get buyer's cash status */
   bcash = player[plid].cash;

   /* check if the buyer's got all the cash to actually buy.. */
   if (((bcash - qty*pr) < 0) && (plid != 0))  {
      buy_pr_remove (cid, plid, 0);
      /*      qty = bcash / pr;  */
      violation (plid);  
      return -1;
   }
   
   /* make sure seller has stuff to sell, else punish her..   */
   if (holding_seller->qty < sqty) {
      sell_pr_remove (cid, plid, 0);
/*      sqty = holding_seller->qty; */
      violation (splid); 
      return -1;
   }
      
   /* check if buyer holds any stocks of same type already */
   if ((holding_buyer = holding_search(plid, cid)) == NULL) 
     holding_buyer = holding_insert (plid, cid, 0);

   /* qty is set to the no. of stocks to actually undergo this transaction */
   if (qty > sqty) {
      /* qty for sale < qty reqd. */
      
      qty = sqty;
      
      if (caller == BUYER)
	sell_pr_remove (cid, splid, 0);
      else 
	comp->buy[0]->qty -= qty;
   }
   else if (qty < sqty) {
      /* qty reqd. < qty for sale */
      
      if (caller == SELLER)
	buy_pr_remove (cid, plid, 0);
      else
	comp->sell[0]->qty -= qty;
   }
   else {
      /* qty reqd. == qty for sale */
      
      if (caller == SELLER)
	buy_pr_remove (cid, plid, 0);
      else
	sell_pr_remove (cid, splid, 0);
   }
   
   cash_changing_hands = qty * pr;
   
   /* set the price at which the trade happened */
   set_last_trade (cid, pr);

   
#ifdef LIFE
   /* lock the player structs */
   pthread_mutex_lock (&player[splid].player_mut);
   if (plid != splid)
     pthread_mutex_lock (&player[plid].player_mut);
   
#endif
   
   /* cash transactions */
   player[plid].cash -= cash_changing_hands;
   player[splid].cash += cash_changing_hands;

   /* stock  transactions*/
   holding_seller->qty -= qty;   
   holding_buyer->qty += qty;
   
#ifdef LIFE
   /* unlock the player structs */
   pthread_mutex_unlock (&player[splid].player_mut);
   if (plid != splid) 
     pthread_mutex_unlock (&player[plid].player_mut);
#endif
   
#ifdef LOG
   fprintf (logfile, "seller : %d -> buyer : %d\t qty : %d\t pr : %d\n",
	   splid, plid, qty, pr);
#endif
   
   if (splid > 0) 
     addto_result_list (splid, POST_SELL_QUOT, cid, qty, pr);
   if (plid > 0)
     addto_result_list (plid, POST_BUY_QUOT, cid, qty, pr);
   return qty;
   /* phew.. */
}

void buy_pr_quote (int client, plid_t plid, cid_t cid, pr_t pr, qty_t qty) {
   qty_t bought = 0, bought_tot = 0;
   pr_t qpr, cpr;
   company_t *c = &company[cid];
   
   while (qty != 0) {
      cpr = c->sell[0]->pr;
      
      if (cpr < pr)
	qpr = pr;
      else if ((cpr - pr) <= tolerance)
	qpr = cpr;
      else {
	 buy_pr_insert (cid, plid, pr, qty);
	 break;
      }
      
      if ((bought = buy_really (plid, cid, qty, qpr, BUYER)) < 0) 
	continue;
      
      bought_tot += bought;
      qty -= bought;
      
   }

   return;
}
	     
void buy_pr_quote_change (int client, plid_t plid, cid_t cid, pr_t pr, 
			  qty_t qty) {

   register int i;
   stock_company_t *temp;
   company_t *c = &company[cid];
   
   for (i=0; i < max_simult_quot; ++i) {
      temp = c->buy[i];
      if ((temp->plid == plid) &&
	  (temp->qty == qty) &&
	  (temp->pr == pr)) {
	   
	 buy_pr_remove (cid, plid, 0);
	 buy_pr_quote (client, plid, cid, pr, qty);
	 break;
      }
   }
}

void sell_pr_quote (int client, plid_t plid, cid_t cid, pr_t pr, qty_t qty) {
   qty_t sold=0, sold_tot = 0;
   pr_t qpr, cpr;
   company_t *c = &company[cid];
   
   while (qty != 0) {
      cpr = c->buy[0]->pr;
      
      if (pr < cpr) 
	qpr = pr;
      else if ((pr - cpr) <= tolerance)
	qpr = cpr;
      else {
	 sell_pr_insert (cid, plid, pr, qty);
	 break;
      }
      
      if ((sold = buy_really (c->buy[0]->plid, cid, qty, qpr, SELLER)) < 0)
	continue;
      
      sold_tot += sold;
      qty -= sold;
      
   }

   return;
}

void sell_pr_quote_change (int client, plid_t plid, cid_t cid, pr_t pr, 
			   qty_t qty) {
   register u_int i;
   stock_company_t *temp;
   company_t *c = &company[cid];
   
   for (i=0; i < max_simult_quot; ++i) {
      temp = c->sell[i];
      if ((temp->plid == plid) &&
	  (temp->qty == qty) &&
	  (temp->pr == pr)) {
	 
	 sell_pr_remove (cid, plid, 0);
	 sell_pr_quote (client, plid, cid, pr, qty);
	 break;
      }
   }
}
